"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest
import yaml

import salt.modules.pkg_resource as pkg_resource
import salt.utils.data
import salt.utils.yaml
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pkg_resource: {}}


def test_pack_sources():
    """
    Test to accepts list of dicts (or a string representing a
    list of dicts) and packs the key/value pairs into a single dict.
    """
    with patch.object(
        salt.utils.yaml,
        "safe_load",
        MagicMock(side_effect=yaml.parser.ParserError("f")),
    ):
        with patch.dict(pkg_resource.__salt__, {"pkg.normalize_name": MagicMock()}):
            assert pkg_resource.pack_sources("sources") == {}

            assert pkg_resource.pack_sources(["A", "a"]) == {}

            assert pkg_resource.pack_sources([{"A": "a"}])


def test_parse_targets():
    """
    Test to parses the input to pkg.install and
    returns back the package(s) to be installed. Returns a
    list of packages, as well as a string noting whether the
    packages are to come from a repository or a binary package.
    """
    with patch.dict(pkg_resource.__grains__, {"os": "A"}):
        assert pkg_resource.parse_targets(pkgs="a", sources="a") == (None, None)

        with patch.object(pkg_resource, "_repack_pkgs", return_value=False):
            assert pkg_resource.parse_targets(pkgs="a") == (None, None)

        with patch.object(pkg_resource, "_repack_pkgs", return_value="A"):
            assert pkg_resource.parse_targets(pkgs="a") == ("A", "repository")

    with patch.dict(pkg_resource.__grains__, {"os": "MacOS1"}):
        with patch.object(pkg_resource, "pack_sources", return_value=False):
            assert pkg_resource.parse_targets(sources="s") == (None, None)

        with patch.object(pkg_resource, "pack_sources", return_value={"A": "/a"}):
            with patch.dict(
                pkg_resource.__salt__,
                {"config.valid_fileproto": MagicMock(return_value=False)},
            ):
                assert pkg_resource.parse_targets(sources="s") == (["/a"], "file")

        with patch.object(pkg_resource, "pack_sources", return_value={"A": "a"}):
            with patch.dict(
                pkg_resource.__salt__,
                {"config.valid_fileproto": MagicMock(return_value=False)},
            ):
                assert pkg_resource.parse_targets(name="n") == (
                    {"n": None},
                    "repository",
                )

                assert pkg_resource.parse_targets() == (None, None)


def test_version():
    """
    Test to Common interface for obtaining the version
    of installed packages.
    """
    with patch.object(salt.utils.data, "is_true", return_value=True):
        mock = MagicMock(return_value={"A": "B"})
        with patch.dict(pkg_resource.__salt__, {"pkg.list_pkgs": mock}):
            assert pkg_resource.version("A") == "B"

            assert pkg_resource.version() == {}

        mock = MagicMock(return_value={})
        with patch.dict(pkg_resource.__salt__, {"pkg.list_pkgs": mock}):
            with patch("builtins.next") as mock_next:
                mock_next.side_effect = StopIteration()
                assert pkg_resource.version("A") == ""


def test_add_pkg():
    """
    Test to add a package to a dict of installed packages.
    """
    assert pkg_resource.add_pkg({"pkgs": []}, "name", "version") is None


def test_sort_pkglist():
    """
    Test to accepts a dict obtained from pkg.list_pkgs() and sorts
    in place the list of versions for any packages that have multiple
    versions installed, so that two package lists can be compared
    to one another.
    """
    assert pkg_resource.sort_pkglist({}) is None


def test_format_pkg_list_no_attr():
    """
    Test to output format of the package list with no attr parameter.
    """
    packages = {
        "glibc": [
            {
                "version": "2.12",
                "epoch": None,
                "release": "1.212.el6",
                "arch": "x86_64",
            }
        ],
        "glibc.i686": [
            {
                "version": "2.12",
                "epoch": None,
                "release": "1.212.el6",
                "arch": "i686",
            }
        ],
        "foobar": [
            {"version": "1.2.0", "epoch": "2", "release": "7", "arch": "x86_64"},
            {"version": "1.2.3", "epoch": "2", "release": "27", "arch": "x86_64"},
        ],
        "foobar.something": [
            {"version": "1.1", "epoch": "3", "release": "23.1", "arch": "i686"}
        ],
        "foobar.": [
            {"version": "1.1", "epoch": "3", "release": "23.1", "arch": "i686"}
        ],
    }
    expected_pkg_list = {
        "glibc": "2.12-1.212.el6",
        "glibc.i686": "2.12-1.212.el6",
        "foobar": "2:1.2.0-7,2:1.2.3-27",
        "foobar.something": "3:1.1-23.1",
        "foobar.": "3:1.1-23.1",
    }
    assert pkg_resource.format_pkg_list(packages, False, None) == expected_pkg_list


def test_format_pkg_list_with_attr():
    """
    Test to output format of the package list with attr parameter.
    In this case, any redundant "arch" reference will be removed
    from the package name since it's included as part of the requested attr.
    """
    name_arch_mapping = {
        "glibc": {"name": "glibc", "arch": None},
        "glibc.i686": {"name": "glibc", "arch": "i686"},
        "foobar": {"name": "foobar", "arch": None},
        "foobar.something": {"name": "foobar.something", "arch": None},
        "foobar.": {"name": "foobar.", "arch": None},
    }
    packages = {
        "glibc": [
            {
                "version": "2.12",
                "epoch": None,
                "release": "1.212.el6",
                "arch": "x86_64",
            }
        ],
        "glibc.i686": [
            {
                "version": "2.12",
                "epoch": None,
                "release": "1.212.el6",
                "arch": "i686",
            }
        ],
        "foobar": [
            {"version": "1.2.0", "epoch": "2", "release": "7", "arch": "x86_64"},
            {"version": "1.2.3", "epoch": "2", "release": "27", "arch": "x86_64"},
        ],
        "foobar.something": [
            {"version": "1.1", "epoch": "3", "release": "23.1", "arch": "i686"}
        ],
        "foobar.": [
            {"version": "1.1", "epoch": "3", "release": "23.1", "arch": "i686"}
        ],
    }
    expected_pkg_list = {
        "glibc": [
            {
                "arch": "x86_64",
                "release": "1.212.el6",
                "epoch": None,
                "version": "2.12",
            },
            {
                "arch": "i686",
                "release": "1.212.el6",
                "epoch": None,
                "version": "2.12",
            },
        ],
        "foobar": [
            {"arch": "x86_64", "release": "7", "epoch": "2", "version": "1.2.0"},
            {"arch": "x86_64", "release": "27", "epoch": "2", "version": "1.2.3"},
        ],
        "foobar.": [
            {"arch": "i686", "release": "23.1", "epoch": "3", "version": "1.1"}
        ],
        "foobar.something": [
            {"arch": "i686", "release": "23.1", "epoch": "3", "version": "1.1"}
        ],
    }
    with patch.dict(pkg_resource.__salt__, {"pkg.parse_arch": name_arch_mapping.get}):
        pkgs = pkg_resource.format_pkg_list(packages, False, attr=["epoch", "release"])
        assert sorted(pkgs) == sorted(expected_pkg_list)


def test_repack_pkgs():
    """
    Test to check that repack function is raising error in case of
    package name collisions
    """
    assert pkg_resource._repack_pkgs([{"A": "a"}])
    assert pkg_resource._repack_pkgs([{"A": "a"}, {"B": "b"}])
    with pytest.raises(SaltInvocationError):
        assert pkg_resource._repack_pkgs([{"A": "a"}, {"A": "c"}])


def test_stringify():
    """
    Test to takes a dict of package name/version information
    and joins each list of
    installed versions into a string.
    """
    assert pkg_resource.stringify({}) is None


def test_version_clean():
    """
    Test to clean the version string removing extra data.
    """
    with patch.dict(
        pkg_resource.__salt__, {"pkg.version_clean": MagicMock(return_value="A")}
    ):
        assert pkg_resource.version_clean("version") == "A"

    assert pkg_resource.version_clean("v") == "v"


def test_check_extra_requirements():
    """
    Test to check if the installed package already
    has the given requirements.
    """
    with patch.dict(
        pkg_resource.__salt__,
        {"pkg.check_extra_requirements": MagicMock(return_value="A")},
    ):
        assert pkg_resource.check_extra_requirements("a", "b") == "A"

    assert pkg_resource.check_extra_requirements("a", False)


def test_version_compare():
    """
    Test the version_compare function

    TODO: Come up with a good way to test epoch handling across different
    platforms. This function will look in the ``__salt__`` dunder for a
    version_cmp function (which not all pkg modules implement) and use that
    to perform platform-specific handling (including interpretation of
    epochs), but even an integration test would need to take into account
    the fact that not all package managers grok epochs.
    """
    assert pkg_resource.version_compare("2.0", "<", "3.0") is True



import logging
import os
import sys

import pytest

import salt.states.pip_state as pip_state
import salt.utils.path
import salt.version
from tests.support.mixins import LoaderModuleMockMixin, SaltReturnAssertsMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

pip = pytest.importorskip(
    "pip", reason="The 'pip' library is not importable(installed system-wide)"
)

log = logging.getLogger(__name__)


class PipStateTest(TestCase, SaltReturnAssertsMixin, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {
            pip_state: {
                "__env__": "base",
                "__opts__": {"test": False},
                "__salt__": {"cmd.which_bin": lambda _: "pip"},
            }
        }

    def test_install_requirements_parsing(self):
        log.debug("Real pip version is %s", pip.__version__)
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        pip_list = MagicMock(return_value={"pep8": "1.3.3"})
        pip_version = pip.__version__
        mock_pip_version = MagicMock(return_value=pip_version)
        with patch.dict(pip_state.__salt__, {"pip.version": mock_pip_version}):
            with patch.dict(
                pip_state.__salt__, {"cmd.run_all": mock, "pip.list": pip_list}
            ):
                with patch.dict(pip_state.__opts__, {"test": True}):
                    log.debug(
                        "pip_state._from_line globals: %s",
                        [name for name in pip_state._from_line.__globals__],
                    )
                    ret = pip_state.installed("pep8=1.3.2")
                    self.assertSaltFalseReturn({"test": ret})
                    self.assertInSaltComment(
                        "Invalid version specification in package pep8=1.3.2. "
                        "'=' is not supported, use '==' instead.",
                        {"test": ret},
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"pep8": "1.3.3"})
            pip_install = MagicMock(return_value={"retcode": 0})
            with patch.dict(
                pip_state.__salt__,
                {"cmd.run_all": mock, "pip.list": pip_list, "pip.install": pip_install},
            ):
                with patch.dict(pip_state.__opts__, {"test": True}):
                    ret = pip_state.installed("pep8>=1.3.2")
                    self.assertSaltTrueReturn({"test": ret})
                    self.assertInSaltComment(
                        "Python package pep8>=1.3.2 was already installed",
                        {"test": ret},
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"pep8": "1.3.3"})
            with patch.dict(
                pip_state.__salt__, {"cmd.run_all": mock, "pip.list": pip_list}
            ):
                with patch.dict(pip_state.__opts__, {"test": True}):
                    ret = pip_state.installed("pep8<1.3.2")
                    self.assertSaltNoneReturn({"test": ret})
                    self.assertInSaltComment(
                        "Python package pep8<1.3.2 is set to be installed",
                        {"test": ret},
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"pep8": "1.3.2"})
            pip_install = MagicMock(return_value={"retcode": 0})
            with patch.dict(
                pip_state.__salt__,
                {"cmd.run_all": mock, "pip.list": pip_list, "pip.install": pip_install},
            ):
                with patch.dict(pip_state.__opts__, {"test": True}):
                    ret = pip_state.installed("pep8>1.3.1,<1.3.3")
                    self.assertSaltTrueReturn({"test": ret})
                    self.assertInSaltComment(
                        "Python package pep8>1.3.1,<1.3.3 was already installed",
                        {"test": ret},
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"pep8": "1.3.1"})
            pip_install = MagicMock(return_value={"retcode": 0})
            with patch.dict(
                pip_state.__salt__,
                {"cmd.run_all": mock, "pip.list": pip_list, "pip.install": pip_install},
            ):
                with patch.dict(pip_state.__opts__, {"test": True}):
                    ret = pip_state.installed("pep8>1.3.1,<1.3.3")
                    self.assertSaltNoneReturn({"test": ret})
                    self.assertInSaltComment(
                        "Python package pep8>1.3.1,<1.3.3 is set to be installed",
                        {"test": ret},
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"pep8": "1.3.1"})
            with patch.dict(
                pip_state.__salt__, {"cmd.run_all": mock, "pip.list": pip_list}
            ):
                with patch.dict(pip_state.__opts__, {"test": True}):
                    ret = pip_state.installed(
                        "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting>=0.5.1"
                    )
                    self.assertSaltNoneReturn({"test": ret})
                    self.assertInSaltComment(
                        "Python package git+https://github.com/saltstack/"
                        "salt-testing.git#egg=SaltTesting>=0.5.1 is set to be "
                        "installed",
                        {"test": ret},
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"pep8": "1.3.1"})
            with patch.dict(
                pip_state.__salt__, {"cmd.run_all": mock, "pip.list": pip_list}
            ):
                with patch.dict(pip_state.__opts__, {"test": True}):
                    ret = pip_state.installed(
                        "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting"
                    )
                    self.assertSaltNoneReturn({"test": ret})
                    self.assertInSaltComment(
                        "Python package git+https://github.com/saltstack/"
                        "salt-testing.git#egg=SaltTesting is set to be "
                        "installed",
                        {"test": ret},
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"pep8": "1.3.1"})
            with patch.dict(
                pip_state.__salt__, {"cmd.run_all": mock, "pip.list": pip_list}
            ):
                with patch.dict(pip_state.__opts__, {"test": True}):
                    ret = pip_state.installed(
                        "https://pypi.python.org/packages/source/S/SaltTesting/"
                        "SaltTesting-0.5.0.tar.gz"
                        "#md5=e6760af92b7165f8be53b5763e40bc24"
                    )
                    self.assertSaltNoneReturn({"test": ret})
                    self.assertInSaltComment(
                        "Python package https://pypi.python.org/packages/source/"
                        "S/SaltTesting/SaltTesting-0.5.0.tar.gz"
                        "#md5=e6760af92b7165f8be53b5763e40bc24 is set to be "
                        "installed",
                        {"test": ret},
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"SaltTesting": "0.5.0"})
            pip_install = MagicMock(
                return_value={
                    "retcode": 0,
                    "stderr": "",
                    "stdout": (
                        "Downloading/unpacking https://pypi.python.org/packages"
                        "/source/S/SaltTesting/SaltTesting-0.5.0.tar.gz\n  "
                        "Downloading SaltTesting-0.5.0.tar.gz\n  Running "
                        "setup.py egg_info for package from "
                        "https://pypi.python.org/packages/source/S/SaltTesting/"
                        "SaltTesting-0.5.0.tar.gz\n    \nCleaning up..."
                    ),
                }
            )
            with patch.dict(
                pip_state.__salt__,
                {"cmd.run_all": mock, "pip.list": pip_list, "pip.install": pip_install},
            ):
                ret = pip_state.installed(
                    "https://pypi.python.org/packages/source/S/SaltTesting/"
                    "SaltTesting-0.5.0.tar.gz"
                    "#md5=e6760af92b7165f8be53b5763e40bc24"
                )
                self.assertSaltTrueReturn({"test": ret})
                self.assertInSaltComment(
                    "All packages were successfully installed", {"test": ret}
                )
                self.assertInSaltReturn(
                    "Installed",
                    {"test": ret},
                    (
                        "changes",
                        "https://pypi.python.org/packages/source/S/"
                        "SaltTesting/SaltTesting-0.5.0.tar.gz"
                        "#md5=e6760af92b7165f8be53b5763e40bc24==???",
                    ),
                )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"SaltTesting": "0.5.0"})
            pip_install = MagicMock(
                return_value={"retcode": 0, "stderr": "", "stdout": "Cloned!"}
            )
            with patch.dict(
                pip_state.__salt__,
                {"cmd.run_all": mock, "pip.list": pip_list, "pip.install": pip_install},
            ):
                with patch.dict(pip_state.__opts__, {"test": False}):
                    ret = pip_state.installed(
                        "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting"
                    )
                    self.assertSaltTrueReturn({"test": ret})
                    self.assertInSaltComment(
                        "packages are already installed", {"test": ret}
                    )

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"pep8": "1.3.1"})
            pip_install = MagicMock(return_value={"retcode": 0})
            with patch.dict(
                pip_state.__salt__,
                {"cmd.run_all": mock, "pip.list": pip_list, "pip.install": pip_install},
            ):
                with patch.dict(pip_state.__opts__, {"test": False}):
                    ret = pip_state.installed(
                        "arbitrary ID that should be ignored due to requirements"
                        " specified",
                        requirements="/tmp/non-existing-requirements.txt",
                    )
                    self.assertSaltTrueReturn({"test": ret})

            # Test VCS installations using git+git://
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_list = MagicMock(return_value={"SaltTesting": "0.5.0"})
            pip_install = MagicMock(
                return_value={"retcode": 0, "stderr": "", "stdout": "Cloned!"}
            )
            with patch.dict(
                pip_state.__salt__,
                {"cmd.run_all": mock, "pip.list": pip_list, "pip.install": pip_install},
            ):
                with patch.dict(pip_state.__opts__, {"test": False}):
                    ret = pip_state.installed(
                        "git+git://github.com/saltstack/salt-testing.git#egg=SaltTesting"
                    )
                    self.assertSaltTrueReturn({"test": ret})
                    self.assertInSaltComment(
                        "packages are already installed", {"test": ret}
                    )

    def test_install_requirements_custom_pypi(self):
        """
        test requirement parsing for both when a custom
        pypi index-url is set and when it is not and
        the requirement is already installed.
        """

        # create requirements file
        req_filename = os.path.join(
            RUNTIME_VARS.TMP_STATE_TREE, "custom-pypi-requirements.txt"
        )
        with salt.utils.files.fopen(req_filename, "wb") as reqf:
            reqf.write(b"pep8\n")

        site_pkgs = "/tmp/pip-env/lib/python3.7/site-packages"
        check_stdout = [
            "Looking in indexes: https://custom-pypi-url.org,"
            "https://pypi.org/simple/\nRequirement already satisfied: pep8 in {1}"
            "(from -r /tmp/files/prod/{0} (line 1)) (1.7.1)".format(
                req_filename, site_pkgs
            ),
            "Requirement already satisfied: pep8 in {1}"
            "(from -r /tmp/files/prod/{0} (line1)) (1.7.1)".format(
                req_filename, site_pkgs
            ),
        ]
        pip_version = pip.__version__
        mock_pip_version = MagicMock(return_value=pip_version)

        for stdout in check_stdout:
            pip_install = MagicMock(return_value={"retcode": 0, "stdout": stdout})
            with patch.dict(pip_state.__salt__, {"pip.version": mock_pip_version}):
                with patch.dict(pip_state.__salt__, {"pip.install": pip_install}):
                    ret = pip_state.installed(name="", requirements=req_filename)
                    self.assertSaltTrueReturn({"test": ret})
                    assert "Requirements were already installed." == ret["comment"]

    def test_install_requirements_custom_pypi_changes(self):
        """
        test requirement parsing for both when a custom
        pypi index-url is set and when it is not and
        the requirement is not installed.
        """

        # create requirements file
        req_filename = os.path.join(
            RUNTIME_VARS.TMP_STATE_TREE, "custom-pypi-requirements.txt"
        )
        with salt.utils.files.fopen(req_filename, "wb") as reqf:
            reqf.write(b"pep8\n")

        site_pkgs = "/tmp/pip-env/lib/python3.7/site-packages"
        check_stdout = [
            "Looking in indexes:"
            " https://custom-pypi-url.org,https://pypi.org/simple/\nCollecting pep8\n "
            " Using"
            " cachedhttps://custom-pypi-url.org//packages/42/3f/669429cef5acb4/pep8-1.7.1-py2.py3-none-any.whl"
            " (41 kB)\nInstalling collected packages: pep8\nSuccessfully installed"
            " pep8-1.7.1",
            "Collecting pep8\n  Using"
            " cachedhttps://custom-pypi-url.org//packages/42/3f/669429cef5acb4/pep8-1.7.1-py2.py3-none-any.whl"
            " (41 kB)\nInstalling collected packages: pep8\nSuccessfully installed"
            " pep8-1.7.1",
        ]

        pip_version = pip.__version__
        mock_pip_version = MagicMock(return_value=pip_version)

        for stdout in check_stdout:
            pip_install = MagicMock(return_value={"retcode": 0, "stdout": stdout})
            with patch.dict(pip_state.__salt__, {"pip.version": mock_pip_version}):
                with patch.dict(pip_state.__salt__, {"pip.install": pip_install}):
                    ret = pip_state.installed(name="", requirements=req_filename)
                    self.assertSaltTrueReturn({"test": ret})
                    assert (
                        "Successfully processed requirements file {}.".format(
                            req_filename
                        )
                        == ret["comment"]
                    )

    def test_install_in_editable_mode(self):
        """
        Check that `name` parameter containing bad characters is not parsed by
        pip when package is being installed in editable mode.
        For more information, see issue #21890.
        """
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        pip_list = MagicMock(return_value={})
        pip_install = MagicMock(
            return_value={"retcode": 0, "stderr": "", "stdout": "Cloned!"}
        )
        pip_version = MagicMock(return_value="10.0.1")
        with patch.dict(
            pip_state.__salt__,
            {
                "cmd.run_all": mock,
                "pip.list": pip_list,
                "pip.install": pip_install,
                "pip.version": pip_version,
            },
        ):
            ret = pip_state.installed(
                "state@name", cwd="/path/to/project", editable=["."]
            )
            self.assertSaltTrueReturn({"test": ret})
            self.assertInSaltComment("successfully installed", {"test": ret})

    def test_install_with_specified_user(self):
        """
        Check that if `user` parameter is set and the user does not exists
        it will fail with an error, see #65458
        """
        user_info = MagicMock(return_value={})
        pip_version = MagicMock(return_value="10.0.1")
        with patch.dict(
            pip_state.__salt__,
            {
                "user.info": user_info,
                "pip.version": pip_version,
            },
        ):
            ret = pip_state.installed("mypkg", user="fred")
            self.assertSaltFalseReturn({"test": ret})
            self.assertInSaltComment("User fred does not exist", {"test": ret})


class PipStateUtilsTest(TestCase):
    def test_has_internal_exceptions_mod_function(self):
        assert pip_state.pip_has_internal_exceptions_mod("10.0")
        assert pip_state.pip_has_internal_exceptions_mod("18.1")
        assert not pip_state.pip_has_internal_exceptions_mod("9.99")

    def test_has_exceptions_mod_function(self):
        assert pip_state.pip_has_exceptions_mod("1.0")
        assert not pip_state.pip_has_exceptions_mod("0.1")
        assert not pip_state.pip_has_exceptions_mod("10.0")

    def test_pip_purge_method_with_pip(self):
        mock_modules = sys.modules.copy()
        mock_modules.pop("pip", None)
        mock_modules["pip"] = object()
        with patch("sys.modules", mock_modules):
            pip_state.purge_pip()
        assert "pip" not in mock_modules

    def test_pip_purge_method_without_pip(self):
        mock_modules = sys.modules.copy()
        mock_modules.pop("pip", None)
        with patch("sys.modules", mock_modules):
            pip_state.purge_pip()



import logging
import time

import pytest

from salt.utils.timeout import wait_for
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


def return_something_after(seconds, something=True):
    start = time.time()
    end = start + seconds
    log.debug("Will return %s at %s", something, end)

    def actual():
        t = time.time()
        condition = t >= end
        log.debug("Return something at %s ? %s", t, condition)
        if condition:
            return something
        else:
            return False

    return actual


def return_args_after(seconds):
    start = time.time()
    end = start + seconds

    def actual(*args):
        if time.time() >= end:
            return args
        else:
            return False

    return actual


def return_kwargs_after(seconds):
    start = time.time()
    end = start + seconds

    def actual(**kwargs):
        if time.time() >= end:
            return kwargs
        else:
            return False

    return actual


class WaitForTests(TestCase):
    def setUp(self):
        self.true_after_1s = return_something_after(1)
        self.self_after_1s = return_something_after(1, something=self)

    def tearDown(self):
        del self.true_after_1s
        del self.self_after_1s

    @pytest.mark.slow_test
    def test_wait_for_true(self):
        ret = wait_for(self.true_after_1s, timeout=2, step=0.5)
        self.assertTrue(ret)

    @pytest.mark.slow_test
    def test_wait_for_self(self):
        ret = wait_for(self.self_after_1s, timeout=2, step=0.5)
        self.assertEqual(ret, self)

    def test_wait_for_too_long(self):
        ret = wait_for(self.true_after_1s, timeout=0.5, step=0.1, default=False)
        self.assertFalse(ret)

    @pytest.mark.slow_test
    @pytest.mark.skip_initial_gh_actions_failure
    def test_wait_for_with_big_step(self):
        ret = wait_for(self.true_after_1s, timeout=1.5, step=2)
        self.assertTrue(ret)

    @pytest.mark.slow_test
    def test_wait_for_custom_args(self):
        args_after_1s = return_args_after(1)
        args = ("one", "two")
        ret = wait_for(args_after_1s, timeout=2, step=0.5, func_args=args)
        self.assertEqual(ret, args)

    @pytest.mark.slow_test
    def test_wait_for_custom_kwargs(self):
        kwargs_after_1s = return_kwargs_after(1)
        kwargs = {"one": 1, "two": 2}
        ret = wait_for(kwargs_after_1s, timeout=2, step=0.5, func_kwargs=kwargs)
        self.assertEqual(ret, kwargs)

    def test_return_false(self):
        ret = self.true_after_1s()
        self.assertFalse(ret)



"""
    tests.pytests.unit.beacons.test_load
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Load beacon test cases
"""

import pytest

import salt.beacons.load as load
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {load: {"__context__": {}, "__salt__": {}}}


def test_non_list_config():
    config = {}

    ret = load.validate(config)
    assert ret == (False, "Configuration for load beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = load.validate(config)
    assert ret == (False, "Averages configuration is required for load beacon.")


@pytest.mark.skip_on_windows(reason="os.getloadavg not available on Windows")
def test_load_match():
    with patch("os.getloadavg", MagicMock(return_value=(1.82, 1.84, 1.56))):
        config = [
            {
                "averages": {"1m": [0.0, 2.0], "5m": [0.0, 1.5], "15m": [0.0, 1.0]},
                "emitatstartup": True,
                "onchangeonly": False,
            }
        ]

        ret = load.validate(config)
        assert ret == (True, "Valid beacon configuration")

        _expected_return = [{"1m": 1.82, "5m": 1.84, "15m": 1.56}]
        ret = load.beacon(config)
        assert ret == _expected_return



"""
Connection module for Amazon Autoscale Groups

.. versionadded:: 2014.7.0

:configuration: This module accepts explicit autoscale credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        asg.keyid: GKTADJGHEIQSXMKKRBJ08H
        asg.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        asg.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

:depends: boto
:depends: boto3
"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import datetime
import email.mime.multipart
import logging
import sys
import time
from collections import OrderedDict

import salt.utils.compat
import salt.utils.json
import salt.utils.versions

log = logging.getLogger(__name__)
DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


try:
    import boto
    import boto.ec2
    import boto.ec2.autoscale as autoscale
    import boto.ec2.blockdevicemapping as blockdevicemapping
    import boto.ec2.instance

    logging.getLogger("boto").setLevel(logging.CRITICAL)
    import boto3  # pylint: disable=unused-import
    from botocore.exceptions import ClientError

    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False


def __virtual__():
    """
    Only load if boto libraries exist.
    """
    has_boto_reqs = salt.utils.versions.check_boto_reqs()
    if has_boto_reqs is True:
        __utils__["boto.assign_funcs"](
            __name__, "asg", module="ec2.autoscale", pack=__salt__
        )
        setattr(
            sys.modules[__name__],
            "_get_ec2_conn",
            __utils__["boto.get_connection_func"]("ec2"),
        )
    return has_boto_reqs


def __init__(opts):
    if HAS_BOTO:
        __utils__["boto3.assign_funcs"](
            __name__, "autoscaling", get_conn_funcname="_get_conn_autoscaling_boto3"
        )


def exists(name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if an autoscale group exists.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.exists myasg region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            _conn = conn.get_all_groups(names=[name])
            if _conn:
                return True
            else:
                msg = f"The autoscale group does not exist in region {region}"
                log.debug(msg)
                return False
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            return False


def get_config(name, region=None, key=None, keyid=None, profile=None):
    """
    Get the configuration for an autoscale group.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.get_config myasg region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            asg = conn.get_all_groups(names=[name])
            if asg:
                asg = asg[0]
            else:
                return {}
            ret = OrderedDict()
            attrs = [
                "name",
                "availability_zones",
                "default_cooldown",
                "desired_capacity",
                "health_check_period",
                "health_check_type",
                "launch_config_name",
                "load_balancers",
                "max_size",
                "min_size",
                "placement_group",
                "vpc_zone_identifier",
                "tags",
                "termination_policies",
                "suspended_processes",
            ]
            for attr in attrs:
                # Tags are objects, so we need to turn them into dicts.
                if attr == "tags":
                    _tags = []
                    for tag in asg.tags:
                        _tag = OrderedDict()
                        _tag["key"] = tag.key
                        _tag["value"] = tag.value
                        _tag["propagate_at_launch"] = tag.propagate_at_launch
                        _tags.append(_tag)
                    ret["tags"] = _tags
                # Boto accepts a string or list as input for vpc_zone_identifier,
                # but always returns a comma separated list. We require lists in
                # states.
                elif attr == "vpc_zone_identifier":
                    ret[attr] = getattr(asg, attr).split(",")
                # convert SuspendedProcess objects to names
                elif attr == "suspended_processes":
                    suspended_processes = getattr(asg, attr)
                    ret[attr] = sorted(x.process_name for x in suspended_processes)
                else:
                    ret[attr] = getattr(asg, attr)
            # scaling policies
            policies = conn.get_all_policies(as_group=name)
            ret["scaling_policies"] = []
            for policy in policies:
                ret["scaling_policies"].append(
                    dict(
                        [
                            ("name", policy.name),
                            ("adjustment_type", policy.adjustment_type),
                            ("scaling_adjustment", policy.scaling_adjustment),
                            ("min_adjustment_step", policy.min_adjustment_step),
                            ("cooldown", policy.cooldown),
                        ]
                    )
                )
            # scheduled actions
            actions = conn.get_all_scheduled_actions(as_group=name)
            ret["scheduled_actions"] = {}
            for action in actions:
                end_time = None
                if action.end_time:
                    end_time = action.end_time.isoformat()
                ret["scheduled_actions"][action.name] = dict(
                    [
                        ("min_size", action.min_size),
                        ("max_size", action.max_size),
                        # AWS bug
                        ("desired_capacity", int(action.desired_capacity)),
                        ("start_time", action.start_time.isoformat()),
                        ("end_time", end_time),
                        ("recurrence", action.recurrence),
                    ]
                )
            return ret
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            return {}


def create(
    name,
    launch_config_name,
    availability_zones,
    min_size,
    max_size,
    desired_capacity=None,
    load_balancers=None,
    default_cooldown=None,
    health_check_type=None,
    health_check_period=None,
    placement_group=None,
    vpc_zone_identifier=None,
    tags=None,
    termination_policies=None,
    suspended_processes=None,
    scaling_policies=None,
    scheduled_actions=None,
    region=None,
    notification_arn=None,
    notification_types=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create an autoscale group.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.create myasg mylc '["us-east-1a", "us-east-1e"]' 1 10 load_balancers='["myelb", "myelb2"]' tags='[{"key": "Name", value="myasg", "propagate_at_launch": True}]'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if isinstance(availability_zones, str):
        availability_zones = salt.utils.json.loads(availability_zones)
    if isinstance(load_balancers, str):
        load_balancers = salt.utils.json.loads(load_balancers)
    if isinstance(vpc_zone_identifier, str):
        vpc_zone_identifier = salt.utils.json.loads(vpc_zone_identifier)
    if isinstance(tags, str):
        tags = salt.utils.json.loads(tags)
    # Make a list of tag objects from the dict.
    _tags = []
    if tags:
        for tag in tags:
            try:
                key = tag.get("key")
            except KeyError:
                log.error("Tag missing key.")
                return False
            try:
                value = tag.get("value")
            except KeyError:
                log.error("Tag missing value.")
                return False
            propagate_at_launch = tag.get("propagate_at_launch", False)
            _tag = autoscale.Tag(
                key=key,
                value=value,
                resource_id=name,
                propagate_at_launch=propagate_at_launch,
            )
            _tags.append(_tag)
    if isinstance(termination_policies, str):
        termination_policies = salt.utils.json.loads(termination_policies)
    if isinstance(suspended_processes, str):
        suspended_processes = salt.utils.json.loads(suspended_processes)
    if isinstance(scheduled_actions, str):
        scheduled_actions = salt.utils.json.loads(scheduled_actions)
    retries = 30
    while True:
        try:
            _asg = autoscale.AutoScalingGroup(
                name=name,
                launch_config=launch_config_name,
                availability_zones=availability_zones,
                min_size=min_size,
                max_size=max_size,
                desired_capacity=desired_capacity,
                load_balancers=load_balancers,
                default_cooldown=default_cooldown,
                health_check_type=health_check_type,
                health_check_period=health_check_period,
                placement_group=placement_group,
                tags=_tags,
                vpc_zone_identifier=vpc_zone_identifier,
                termination_policies=termination_policies,
                suspended_processes=suspended_processes,
            )
            conn.create_auto_scaling_group(_asg)
            # create scaling policies
            _create_scaling_policies(conn, name, scaling_policies)
            # create scheduled actions
            _create_scheduled_actions(conn, name, scheduled_actions)
            # create notifications
            if notification_arn and notification_types:
                conn.put_notification_configuration(
                    _asg, notification_arn, notification_types
                )
            log.info("Created ASG %s", name)
            return True
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            msg = "Failed to create ASG %s", name
            log.error(msg)
            return False


def update(
    name,
    launch_config_name,
    availability_zones,
    min_size,
    max_size,
    desired_capacity=None,
    load_balancers=None,
    default_cooldown=None,
    health_check_type=None,
    health_check_period=None,
    placement_group=None,
    vpc_zone_identifier=None,
    tags=None,
    termination_policies=None,
    suspended_processes=None,
    scaling_policies=None,
    scheduled_actions=None,
    notification_arn=None,
    notification_types=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Update an autoscale group.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.update myasg mylc '["us-east-1a", "us-east-1e"]' 1 10 load_balancers='["myelb", "myelb2"]' tags='[{"key": "Name", value="myasg", "propagate_at_launch": True}]'
    """

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    conn3 = _get_conn_autoscaling_boto3(
        region=region, key=key, keyid=keyid, profile=profile
    )
    if not conn:
        return False, "failed to connect to AWS"
    if isinstance(availability_zones, str):
        availability_zones = salt.utils.json.loads(availability_zones)
    if isinstance(load_balancers, str):
        load_balancers = salt.utils.json.loads(load_balancers)
    if isinstance(vpc_zone_identifier, str):
        vpc_zone_identifier = salt.utils.json.loads(vpc_zone_identifier)
    if isinstance(tags, str):
        tags = salt.utils.json.loads(tags)
    if isinstance(termination_policies, str):
        termination_policies = salt.utils.json.loads(termination_policies)
    if isinstance(suspended_processes, str):
        suspended_processes = salt.utils.json.loads(suspended_processes)
    if isinstance(scheduled_actions, str):
        scheduled_actions = salt.utils.json.loads(scheduled_actions)

    # Massage our tagset into  add / remove lists
    # Use a boto3 call here b/c the boto2 call doeesn't implement filters
    current_tags = conn3.describe_tags(
        Filters=[{"Name": "auto-scaling-group", "Values": [name]}]
    ).get("Tags", [])
    current_tags = [
        {
            "key": t["Key"],
            "value": t["Value"],
            "resource_id": t["ResourceId"],
            "propagate_at_launch": t.get("PropagateAtLaunch", False),
        }
        for t in current_tags
    ]
    add_tags = []
    desired_tags = []
    if tags:
        tags = __utils__["boto3.ordered"](tags)
        for tag in tags:
            try:
                key = tag.get("key")
            except KeyError:
                log.error("Tag missing key.")
                return False, f"Tag {tag} missing key"
            try:
                value = tag.get("value")
            except KeyError:
                log.error("Tag missing value.")
                return False, f"Tag {tag} missing value"
            propagate_at_launch = tag.get("propagate_at_launch", False)
            _tag = {
                "key": key,
                "value": value,
                "resource_id": name,
                "propagate_at_launch": propagate_at_launch,
            }
            if _tag not in current_tags:
                add_tags.append(_tag)
            desired_tags.append(_tag)
    delete_tags = [t for t in current_tags if t not in desired_tags]

    retries = 30
    while True:
        try:
            _asg = autoscale.AutoScalingGroup(
                connection=conn,
                name=name,
                launch_config=launch_config_name,
                availability_zones=availability_zones,
                min_size=min_size,
                max_size=max_size,
                desired_capacity=desired_capacity,
                load_balancers=load_balancers,
                default_cooldown=default_cooldown,
                health_check_type=health_check_type,
                health_check_period=health_check_period,
                placement_group=placement_group,
                tags=add_tags,
                vpc_zone_identifier=vpc_zone_identifier,
                termination_policies=termination_policies,
            )
            if notification_arn and notification_types:
                conn.put_notification_configuration(
                    _asg, notification_arn, notification_types
                )
            _asg.update()
            # Seems the update call doesn't handle tags, so we'll need to update
            # that separately.
            if add_tags:
                log.debug("Adding/updating tags from ASG: %s", add_tags)
                conn.create_or_update_tags([autoscale.Tag(**t) for t in add_tags])
            if delete_tags:
                log.debug("Deleting tags from ASG: %s", delete_tags)
                conn.delete_tags([autoscale.Tag(**t) for t in delete_tags])
            # update doesn't handle suspended_processes either
            # Resume all processes
            _asg.resume_processes()
            # suspend any that are specified.  Note that the boto default of empty
            # list suspends all; don't do that.
            if suspended_processes:
                _asg.suspend_processes(suspended_processes)
            log.info("Updated ASG %s", name)
            # ### scaling policies
            # delete all policies, then recreate them
            for policy in conn.get_all_policies(as_group=name):
                conn.delete_policy(policy.name, autoscale_group=name)
            _create_scaling_policies(conn, name, scaling_policies)
            # ### scheduled actions
            # delete all scheduled actions, then recreate them
            for scheduled_action in conn.get_all_scheduled_actions(as_group=name):
                conn.delete_scheduled_action(
                    scheduled_action.name, autoscale_group=name
                )
            _create_scheduled_actions(conn, name, scheduled_actions)
            return True, ""
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            msg = f"Failed to update ASG {name}"
            log.error(msg)
            return False, str(e)


def _create_scaling_policies(conn, as_name, scaling_policies):
    "helper function to create scaling policies"
    if scaling_policies:
        for policy in scaling_policies:
            policy = autoscale.policy.ScalingPolicy(
                name=policy["name"],
                as_name=as_name,
                adjustment_type=policy["adjustment_type"],
                scaling_adjustment=policy["scaling_adjustment"],
                min_adjustment_step=policy.get("min_adjustment_step", None),
                cooldown=policy["cooldown"],
            )
            conn.create_scaling_policy(policy)


def _create_scheduled_actions(conn, as_name, scheduled_actions):
    """
    Helper function to create scheduled actions
    """
    if scheduled_actions:
        for name, action in scheduled_actions.items():
            if "start_time" in action and isinstance(action["start_time"], str):
                action["start_time"] = datetime.datetime.strptime(
                    action["start_time"], DATE_FORMAT
                )
            if "end_time" in action and isinstance(action["end_time"], str):
                action["end_time"] = datetime.datetime.strptime(
                    action["end_time"], DATE_FORMAT
                )
            conn.create_scheduled_group_action(
                as_name,
                name,
                desired_capacity=action.get("desired_capacity"),
                min_size=action.get("min_size"),
                max_size=action.get("max_size"),
                start_time=action.get("start_time"),
                end_time=action.get("end_time"),
                recurrence=action.get("recurrence"),
            )


def delete(name, force=False, region=None, key=None, keyid=None, profile=None):
    """
    Delete an autoscale group.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.delete myasg region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            conn.delete_auto_scaling_group(name, force)
            msg = f"Deleted autoscale group {name}."
            log.info(msg)
            return True
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            msg = f"Failed to delete autoscale group {name}"
            log.error(msg)
            return False


def get_cloud_init_mime(cloud_init):
    """
    Get a mime multipart encoded string from a cloud-init dict. Currently
    supports boothooks, scripts and cloud-config.

    CLI Example:

    .. code-block:: bash

        salt myminion boto.get_cloud_init_mime <cloud init>
    """
    if isinstance(cloud_init, str):
        cloud_init = salt.utils.json.loads(cloud_init)
    _cloud_init = email.mime.multipart.MIMEMultipart()
    if "boothooks" in cloud_init:
        for script_name, script in cloud_init["boothooks"].items():
            _script = email.mime.text.MIMEText(script, "cloud-boothook")
            _cloud_init.attach(_script)
    if "scripts" in cloud_init:
        for script_name, script in cloud_init["scripts"].items():
            _script = email.mime.text.MIMEText(script, "x-shellscript")
            _cloud_init.attach(_script)
    if "cloud-config" in cloud_init:
        cloud_config = cloud_init["cloud-config"]
        _cloud_config = email.mime.text.MIMEText(
            salt.utils.yaml.safe_dump(cloud_config, default_flow_style=False),
            "cloud-config",
        )
        _cloud_init.attach(_cloud_config)
    return _cloud_init.as_string()


def launch_configuration_exists(name, region=None, key=None, keyid=None, profile=None):
    """
    Check for a launch configuration's existence.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.launch_configuration_exists mylc
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            lc = conn.get_all_launch_configurations(names=[name])
            if lc:
                return True
            else:
                msg = "The launch configuration does not exist in region {}".format(
                    region
                )
                log.debug(msg)
                return False
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            return False


def get_all_launch_configurations(region=None, key=None, keyid=None, profile=None):
    """
    Fetch and return all Launch Configuration with details.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.get_all_launch_configurations
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            return conn.get_all_launch_configurations()
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            return []


def list_launch_configurations(region=None, key=None, keyid=None, profile=None):
    """
    List all Launch Configurations.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.list_launch_configurations
    """
    ret = get_all_launch_configurations(region, key, keyid, profile)
    return [r.name for r in ret]


def describe_launch_configuration(
    name, region=None, key=None, keyid=None, profile=None
):
    """
    Dump details of a given launch configuration.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.describe_launch_configuration mylc
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            lc = conn.get_all_launch_configurations(names=[name])
            if lc:
                return lc[0]
            else:
                msg = "The launch configuration does not exist in region {}".format(
                    region
                )
                log.debug(msg)
                return None
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            return None


def create_launch_configuration(
    name,
    image_id,
    key_name=None,
    vpc_id=None,
    vpc_name=None,
    security_groups=None,
    user_data=None,
    instance_type="m1.small",
    kernel_id=None,
    ramdisk_id=None,
    block_device_mappings=None,
    instance_monitoring=False,
    spot_price=None,
    instance_profile_name=None,
    ebs_optimized=False,
    associate_public_ip_address=None,
    volume_type=None,
    delete_on_termination=True,
    iops=None,
    use_block_device_types=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create a launch configuration.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.create_launch_configuration mylc image_id=ami-0b9c9f62 key_name='mykey' security_groups='["mygroup"]' instance_type='c3.2xlarge'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if isinstance(security_groups, str):
        security_groups = salt.utils.json.loads(security_groups)
    if isinstance(block_device_mappings, str):
        block_device_mappings = salt.utils.json.loads(block_device_mappings)
    _bdms = []
    if block_device_mappings:
        # Boto requires objects for the mappings and the devices.
        _block_device_map = blockdevicemapping.BlockDeviceMapping()
        for block_device_dict in block_device_mappings:
            for block_device, attributes in block_device_dict.items():
                _block_device = blockdevicemapping.EBSBlockDeviceType()
                for attribute, value in attributes.items():
                    setattr(_block_device, attribute, value)
                _block_device_map[block_device] = _block_device
        _bdms = [_block_device_map]

    # If a VPC is specified, then determine the secgroup id's within that VPC, not
    # within the default VPC. If a security group id is already part of the list,
    # convert_to_group_ids leaves that entry without attempting a lookup on it.
    if security_groups and (vpc_id or vpc_name):
        security_groups = __salt__["boto_secgroup.convert_to_group_ids"](
            security_groups,
            vpc_id=vpc_id,
            vpc_name=vpc_name,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
    lc = autoscale.LaunchConfiguration(
        name=name,
        image_id=image_id,
        key_name=key_name,
        security_groups=security_groups,
        user_data=user_data,
        instance_type=instance_type,
        kernel_id=kernel_id,
        ramdisk_id=ramdisk_id,
        block_device_mappings=_bdms,
        instance_monitoring=instance_monitoring,
        spot_price=spot_price,
        instance_profile_name=instance_profile_name,
        ebs_optimized=ebs_optimized,
        associate_public_ip_address=associate_public_ip_address,
        volume_type=volume_type,
        delete_on_termination=delete_on_termination,
        iops=iops,
        use_block_device_types=use_block_device_types,
    )
    retries = 30
    while True:
        try:
            conn.create_launch_configuration(lc)
            log.info("Created LC %s", name)
            return True
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            msg = f"Failed to create LC {name}"
            log.error(msg)
            return False


def delete_launch_configuration(name, region=None, key=None, keyid=None, profile=None):
    """
    Delete a launch configuration.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_asg.delete_launch_configuration mylc
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            conn.delete_launch_configuration(name)
            log.info("Deleted LC %s", name)
            return True
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            msg = f"Failed to delete LC {name}"
            log.error(msg)
            return False


def get_scaling_policy_arn(
    as_group, scaling_policy_name, region=None, key=None, keyid=None, profile=None
):
    """
    Return the arn for a scaling policy in a specific autoscale group or None
    if not found. Mainly used as a helper method for boto_cloudwatch_alarm, for
    linking alarms to scaling policies.

    CLI Example:

    .. code-block:: bash

        salt '*' boto_asg.get_scaling_policy_arn mygroup mypolicy
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while retries > 0:
        retries -= 1
        try:
            policies = conn.get_all_policies(as_group=as_group)
            for policy in policies:
                if policy.name == scaling_policy_name:
                    return policy.policy_arn
            log.error("Could not convert: %s", as_group)
            return None
        except boto.exception.BotoServerError as e:
            if e.error_code != "Throttling":
                raise
            log.debug("Throttled by API, will retry in 5 seconds")
            time.sleep(5)

    log.error("Maximum number of retries exceeded")
    return None


def get_all_groups(region=None, key=None, keyid=None, profile=None):
    """
    Return all AutoScale Groups visible in the account
    (as a list of boto.ec2.autoscale.group.AutoScalingGroup).

    .. versionadded:: 2016.11.0

    CLI Example:

    .. code-block:: bash

        salt-call boto_asg.get_all_groups region=us-east-1 --output yaml

    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            next_token = ""
            asgs = []
            while next_token is not None:
                ret = conn.get_all_groups(next_token=next_token)
                asgs += [a for a in ret]
                next_token = ret.next_token
            return asgs
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            return []


def list_groups(region=None, key=None, keyid=None, profile=None):
    """
    Return all AutoScale Groups visible in the account
    (as a list of names).

    .. versionadded:: 2016.11.0

    CLI Example:

    .. code-block:: bash

        salt-call boto_asg.list_groups region=us-east-1

    """
    return [
        a.name
        for a in get_all_groups(region=region, key=key, keyid=keyid, profile=profile)
    ]


def get_instances(
    name,
    lifecycle_state="InService",
    health_status="Healthy",
    attribute="private_ip_address",
    attributes=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    return attribute of all instances in the named autoscale group.

    CLI Example:

    .. code-block:: bash

        salt-call boto_asg.get_instances my_autoscale_group_name

    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    ec2_conn = _get_ec2_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30
    while True:
        try:
            asgs = conn.get_all_groups(names=[name])
            break
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, retrying in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error(e)
            return False
    if len(asgs) != 1:
        log.debug(
            "name '%s' returns multiple ASGs: %s", name, [asg.name for asg in asgs]
        )
        return False
    asg = asgs[0]
    instance_ids = []
    # match lifecycle_state and health_status
    for i in asg.instances:
        if lifecycle_state is not None and i.lifecycle_state != lifecycle_state:
            continue
        if health_status is not None and i.health_status != health_status:
            continue
        instance_ids.append(i.instance_id)
    # get full instance info, so that we can return the attribute
    instances = ec2_conn.get_only_instances(instance_ids=instance_ids)
    if attributes:
        return [
            [_convert_attribute(instance, attr) for attr in attributes]
            for instance in instances
        ]
    else:
        # properly handle case when not all instances have the requested attribute
        return [
            _convert_attribute(instance, attribute)
            for instance in instances
            if getattr(instance, attribute)
        ]


def _convert_attribute(instance, attribute):
    if attribute == "tags":
        tags = dict(getattr(instance, attribute))
        return {
            key.encode("utf-8"): value.encode("utf-8") for key, value in tags.items()
        }

    return getattr(instance, attribute).encode("ascii")


def enter_standby(
    name,
    instance_ids,
    should_decrement_desired_capacity=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Switch desired instances to StandBy mode

    .. versionadded:: 2016.11.0

    CLI Example:

    .. code-block:: bash

        salt-call boto_asg.enter_standby my_autoscale_group_name '["i-xxxxxx"]'

    """
    conn = _get_conn_autoscaling_boto3(
        region=region, key=key, keyid=keyid, profile=profile
    )
    try:
        response = conn.enter_standby(
            InstanceIds=instance_ids,
            AutoScalingGroupName=name,
            ShouldDecrementDesiredCapacity=should_decrement_desired_capacity,
        )
    except ClientError as e:
        err = __utils__["boto3.get_error"](e)
        if e.response.get("Error", {}).get("Code") == "ResourceNotFoundException":
            return {"exists": False}
        return {"error": err}
    return all(
        activity["StatusCode"] != "Failed" for activity in response["Activities"]
    )


def exit_standby(
    name,
    instance_ids,
    should_decrement_desired_capacity=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Exit desired instances from StandBy mode

    .. versionadded:: 2016.11.0

    CLI Example:

    .. code-block:: bash

        salt-call boto_asg.exit_standby my_autoscale_group_name '["i-xxxxxx"]'

    """
    conn = _get_conn_autoscaling_boto3(
        region=region, key=key, keyid=keyid, profile=profile
    )
    try:
        response = conn.exit_standby(
            InstanceIds=instance_ids, AutoScalingGroupName=name
        )
    except ClientError as e:
        err = __utils__["boto3.get_error"](e)
        if e.response.get("Error", {}).get("Code") == "ResourceNotFoundException":
            return {"exists": False}
        return {"error": err}
    return all(
        activity["StatusCode"] != "Failed" for activity in response["Activities"]
    )



import logging

import pytest
from saltfactories.utils import random_string

import salt.modules.cmdmod as cmdmod
import salt.modules.win_useradd as win_useradd
from tests.support.mock import MagicMock, patch

try:
    import win32net

    WINAPI = True
except ImportError:
    WINAPI = False

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {
        win_useradd: {
            "__salt__": {
                "cmd.run_all": cmdmod.run_all,
            },
        }
    }


@pytest.fixture
def username():
    _username = random_string("test-account-", uppercase=False)
    try:
        yield _username
    finally:
        try:
            win_useradd.delete(_username, purge=True, force=True)
        except Exception:  # pylint: disable=broad-except
            # The point here is just system cleanup. It can fail if no account
            # was created
            pass


@pytest.fixture
def account(username):
    with pytest.helpers.create_account(username=username) as account:
        win_useradd.addgroup(account.username, "Users")
        yield account


@pytest.mark.skipif(not WINAPI, reason="pywin32 not available")
def test_info(account, caplog):
    with caplog.at_level(logging.DEBUG):
        win_useradd.info(account.username)
    assert f"user_name: {account.username}" in caplog.text
    assert "domain_name: ." in caplog.text


@pytest.mark.skipif(not WINAPI, reason="pywin32 not available")
def test_info_domain(account, caplog):
    domain = "mydomain"
    dc = "myDC"
    with caplog.at_level(logging.DEBUG), patch(
        "win32net.NetGetAnyDCName", MagicMock(return_value=dc)
    ):
        account.username = f"{domain}\\{account.username}"
        win_useradd.info(account.username)
    assert f"Found DC: {dc}" in caplog.text


@pytest.mark.skipif(not WINAPI, reason="pywin32 not available")
def test_info_error(account, caplog):
    domain = "mydomain"
    dc = "myDC"
    with caplog.at_level(logging.DEBUG), patch(
        "win32net.NetGetAnyDCName", MagicMock(side_effect=win32net.error)
    ):
        account.username = f"{domain}\\{account.username}"
        win_useradd.info(account.username)
    assert f"DC not found. Using username: {account.username}" in caplog.text



"""
Interface to Red Hat tuned-adm module

:maintainer:    Syed Ali <alicsyed@gmail.com>
:maturity:      new
:depends:       tuned-adm
:platform:      Linux
"""

import re

import salt.utils.path

__func_alias__ = {
    "list_": "list",
}

__virtualname__ = "tuned"


def __virtual__():
    """
    Check to see if tuned-adm binary is installed on the system

    """
    tuned_adm = salt.utils.path.which("tuned-adm")
    if not tuned_adm:
        return (
            False,
            "The tuned execution module failed to load: the tuned-adm binary is not in"
            " the path.",
        )
    return __virtualname__


def list_():
    """
    List the profiles available

    CLI Example:

    .. code-block:: bash

        salt '*' tuned.list
    """

    result = __salt__["cmd.run"]("tuned-adm list").splitlines()
    # Remove "Available profiles:"
    result.pop(0)
    # Cut off any warnings
    try:
        result = result[: result.index("** COLLECTED WARNINGS **") - 1]
    except ValueError:
        pass
    # Remove "Current active profile:.*"
    result.pop()
    # Output can be : " - <profile name> - <description>" (v2.7.1)
    # or " - <profile name> " (v2.4.1)
    result = [i.split("- ")[1].strip() for i in result]
    return result


def active():
    """
    Return current active profile

    CLI Example:

    .. code-block:: bash

        salt '*' tuned.active
    """

    # turn off all profiles
    result = __salt__["cmd.run_all"]("tuned-adm active", ignore_retcode=True)
    if result["retcode"] != 0:
        return "none"
    pattern = re.compile(r"""(?P<stmt>Current active profile:) (?P<profile>\w+.*)""")
    match = re.match(pattern, result["stdout"])
    return "{}".format(match.group("profile"))


def off():
    """
    Turn off all profiles

    CLI Example:

    .. code-block:: bash

        salt '*' tuned.off
    """

    # turn off all profiles
    result = __salt__["cmd.retcode"]("tuned-adm off")
    if int(result) != 0:
        return False
    return True


def profile(profile_name):
    """
    Activate specified profile

    CLI Example:

    .. code-block:: bash

        salt '*' tuned.profile virtual-guest
    """

    # run tuned-adm with the profile specified
    result = __salt__["cmd.retcode"](f"tuned-adm profile {profile_name}")
    if int(result) != 0:
        return False
    return f"{profile_name}"



import ntpath
import os
import platform
import posixpath
import sys
import tempfile
import types

import pytest

import salt.utils.compat
import salt.utils.path
import salt.utils.platform
from salt.exceptions import CommandNotFoundError
from tests.support.mock import patch
from tests.support.unit import TestCase


class PathJoinTestCase(TestCase):

    PLATFORM_FUNC = platform.system
    BUILTIN_MODULES = sys.builtin_module_names

    NIX_PATHS = (
        (("/", "key"), "/key"),
        (("/etc/salt", "/etc/salt/pki"), "/etc/salt/etc/salt/pki"),
        (("/usr/local", "/etc/salt/pki"), "/usr/local/etc/salt/pki"),
    )

    WIN_PATHS = (
        (("c:", "temp", "foo"), "c:\\temp\\foo"),
        (("c:", r"\temp", r"\foo"), "c:\\temp\\foo"),
        (("c:\\", r"\temp", r"\foo"), "c:\\temp\\foo"),
        ((r"c:\\", r"\temp", r"\foo"), "c:\\temp\\foo"),
        (("c:", r"\temp", r"\foo", "bar"), "c:\\temp\\foo\\bar"),
        (("c:", r"\temp", r"\foo\bar"), "c:\\temp\\foo\\bar"),
    )

    @pytest.mark.skip(reason="Skipped until properly mocked")
    @pytest.mark.skip_on_windows
    def test_nix_paths(self):
        for idx, (parts, expected) in enumerate(self.NIX_PATHS):
            path = salt.utils.path.join(*parts)
            assert f"{idx}: {path}" == f"{idx}: {expected}"

    @pytest.mark.skip(reason="Skipped until properly mocked")
    @pytest.mark.skip_unless_on_windows
    def test_windows_paths(self):
        for idx, (parts, expected) in enumerate(self.WIN_PATHS):
            path = salt.utils.path.join(*parts)
            assert f"{idx}: {path}" == f"{idx}: {expected}"

    @pytest.mark.skip(reason="Skipped until properly mocked")
    @pytest.mark.skip_on_windows
    def test_windows_paths_patched_path_module(self):
        self.__patch_path()

        try:
            for idx, (parts, expected) in enumerate(self.WIN_PATHS):
                path = salt.utils.path.join(*parts)
                assert f"{idx}: {path}" == f"{idx}: {expected}"
        finally:
            self.__unpatch_path()

    @pytest.mark.skip_on_windows(reason="*nix-only test")
    def test_mixed_unicode_and_binary(self):
        """
        This tests joining paths that contain a mix of components with unicode
        strings and non-unicode strings with the unicode characters as binary.

        This is no longer something we need to concern ourselves with in
        Python 3, but the test should nonetheless pass on Python 3. Really what
        we're testing here is that we don't get a UnicodeDecodeError when
        running on Python 2.
        """
        a = "/foo/bar"
        b = "Д"
        expected = "/foo/bar/\u0414"
        actual = salt.utils.path.join(a, b)
        assert actual == expected

    def __patch_path(self):
        modules = list(self.BUILTIN_MODULES[:])
        modules.pop(modules.index("posix"))
        modules.append("nt")

        code = """'''Salt unittest loaded NT module'''"""
        module = types.ModuleType("nt")
        exec(code, module.__dict__)
        sys.modules["nt"] = module

        sys.builtin_module_names = modules
        platform.system = lambda: "windows"

        for module in (ntpath, os, os.path, tempfile):
            salt.utils.compat.reload(module)

    def __unpatch_path(self):
        del sys.modules["nt"]
        sys.builtin_module_names = self.BUILTIN_MODULES[:]
        platform.system = self.PLATFORM_FUNC

        for module in (posixpath, os, os.path, tempfile, platform):
            salt.utils.compat.reload(module)


class PathTestCase(TestCase):
    def test_which_bin(self):
        ret = salt.utils.path.which_bin("str")
        assert ret is None

        test_exes = ["ls", "echo"]
        with patch("salt.utils.path.which", return_value="/tmp/dummy_path"):
            ret = salt.utils.path.which_bin(test_exes)
            assert ret == "/tmp/dummy_path"

            ret = salt.utils.path.which_bin([])
            assert ret is None

        with patch("salt.utils.path.which", return_value=""):
            ret = salt.utils.path.which_bin(test_exes)
            assert ret is None

    def test_sanitize_win_path(self):
        assert (
            salt.utils.path.sanitize_win_path("\\windows\\system")
            == "\\windows\\system"
        )
        assert (
            salt.utils.path.sanitize_win_path("\\bo:g|us\\p?at*h>")
            == "\\bo_g_us\\p_at_h_"
        )

    def test_check_or_die(self):
        with pytest.raises(CommandNotFoundError):
            salt.utils.path.check_or_die(None)

        with patch("salt.utils.path.which", return_value=False):
            with pytest.raises(CommandNotFoundError):
                salt.utils.path.check_or_die("FAKE COMMAND")

    def test_join(self):
        with patch(
            "salt.utils.platform.is_windows", return_value=False
        ) as is_windows_mock:
            assert not is_windows_mock.return_value
            expected_path = os.path.join(os.sep + "a", "b", "c", "d")
            ret = salt.utils.path.join("/a/b/c", "d")
            assert ret == expected_path


class TestWhich(TestCase):
    """
    Tests salt.utils.path.which function to ensure that it returns True as
    expected.
    """

    # The mock patch below will make sure that ALL calls to the which function
    # returns None
    def test_missing_binary_in_linux(self):
        # salt.utils.path.which uses platform.is_windows to determine the platform, so we're using linux here
        with patch("salt.utils.platform.is_windows", lambda: False):
            with patch("salt.utils.path.which", lambda exe: None):
                assert salt.utils.path.which("this-binary-does-not-exist") is None

    # The mock patch below will make sure that ALL calls to the which function
    # return whatever is sent to it
    def test_existing_binary_in_linux(self):
        # salt.utils.path.which uses platform.is_windows to determine the platform, so we're using linux here
        with patch("salt.utils.platform.is_windows", lambda: False):
            with patch("salt.utils.path.which", lambda exe: exe):
                assert salt.utils.path.which("this-binary-exists-under-linux")

    def test_existing_binary_in_windows(self):
        with patch("os.path.isfile") as isfile:
            # We define the side_effect attribute on the mocked object in order to
            # specify which calls return which values. First call to os.path.isfile
            # returns X, the second Y, the third Z, etc...
            isfile.side_effect = [
                # The first os.path.isfile should return False due to checking the explicit path (first is_executable)
                False,
                # We will now also return False once so we get a .EXE back from
                # the function, see PATHEXT below.
                False,
                # Lastly return True, this is the windows check.
                True,
            ]

            # Patch os.access so that it always returns True
            with patch("os.access", lambda path, mode: True):
                # Disable os.path.islink
                with patch("os.path.islink", lambda path: False):
                    # we're using ';' as os.pathsep in this test
                    with patch("os.pathsep", ";"):
                        # Let's patch os.environ to provide a custom PATH variable
                        with patch.dict(
                            os.environ,
                            {"PATH": os.sep + "bin", "PATHEXT": ".COM;.EXE;.BAT;.CMD"},
                        ):
                            # Let's also patch is_windows to return True
                            with patch("salt.utils.platform.is_windows", lambda: True):
                                assert salt.utils.path.which(
                                    "this-binary-exists-under-windows"
                                ) == os.path.join(
                                    os.sep + "bin",
                                    "this-binary-exists-under-windows.EXE",
                                )

    def test_missing_binary_in_windows(self):
        with patch("os.access") as osaccess:
            osaccess.side_effect = [
                # The first os.access should return False due to checking the explicit path (first is_executable)
                False,
                # The second, iterating through $PATH, should also return False,
                # still checking for Linux
                # which() will add 4 extra paths to the given one, os.access will
                # be called 5 times
                False,
                False,
                False,
                False,
                False,
            ]
            # we're using ';' as os.pathsep in this test
            with patch("os.pathsep", ";"):
                # Let's patch os.environ to provide a custom PATH variable
                with patch.dict(os.environ, {"PATH": os.sep + "bin"}):
                    # Let's also patch is_widows to return True
                    with patch("salt.utils.platform.is_windows", lambda: True):
                        assert (
                            salt.utils.path.which(
                                "this-binary-is-missing-in-windows.exe"
                            )
                            is None
                        )

    def test_existing_binary_in_windows_pathext(self):
        with patch("os.path.isfile") as isfile:
            # We define the side_effect attribute on the mocked object in order to
            # specify which calls return which values. First call to os.path.isfile
            # returns X, the second Y, the third Z, etc...
            isfile.side_effect = [
                # The first os.path.isfile should return False due to checking the explicit path (first is_executable)
                False,
                # We will now also return False 3 times so we get a .CMD back from
                # the function, see PATHEXT below.
                # Lastly return True, this is the windows check.
                False,
                False,
                False,
                True,
            ]

            # Patch os.access so that it always returns True
            with patch("os.access", lambda path, mode: True):

                # Disable os.path.islink
                with patch("os.path.islink", lambda path: False):

                    # we're using ';' as os.pathsep in this test
                    with patch("os.pathsep", ";"):

                        # Let's patch os.environ to provide a custom PATH variable
                        with patch.dict(
                            os.environ,
                            {
                                "PATH": os.sep + "bin",
                                "PATHEXT": (
                                    ".COM;.EXE;.BAT;.CMD;.VBS;"
                                    ".VBE;.JS;.JSE;.WSF;.WSH;.MSC;.PY"
                                ),
                            },
                        ):

                            # Let's also patch is_windows to return True
                            with patch("salt.utils.platform.is_windows", lambda: True):
                                assert salt.utils.path.which(
                                    "this-binary-exists-under-windows"
                                ) == os.path.join(
                                    os.sep + "bin",
                                    "this-binary-exists-under-windows.CMD",
                                )



# The pam components have been modified to be salty and have been taken from
# the pam module under this licence:
# (c) 2007 Chris AtLee <chris@atlee.ca>
# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license.php
"""
Authenticate against PAM

Provides an authenticate function that will allow the caller to authenticate
a user against the Pluggable Authentication Modules (PAM) on the system.

Implemented using ctypes, so no compilation is necessary.

There is one extra configuration option for pam.  The `pam_service` that is
authenticated against.  This defaults to `login`

.. code-block:: yaml

    auth.pam.service: login

.. note:: Solaris-like (SmartOS, OmniOS, ...) systems may need ``auth.pam.service`` set to ``other``.

.. note:: PAM authentication will not work for the ``root`` user.

    The Python interface to PAM does not support authenticating as ``root``.

.. note:: This module executes itself in a subprocess in order to user the system python
    and pam libraries. We do this to avoid openssl version conflicts when
    running under a salt onedir build.
"""

import logging
import os
import pathlib
import subprocess
import sys
from ctypes import (
    CDLL,
    CFUNCTYPE,
    POINTER,
    Structure,
    c_char,
    c_char_p,
    c_int,
    c_uint,
    c_void_p,
    cast,
    pointer,
    sizeof,
)
from ctypes.util import find_library

HAS_USER = True
try:
    import salt.utils.user
except ImportError:
    HAS_USER = False

log = logging.getLogger(__name__)

try:
    LIBC = CDLL(find_library("c"))

    CALLOC = LIBC.calloc
    CALLOC.restype = c_void_p
    CALLOC.argtypes = [c_uint, c_uint]

    STRDUP = LIBC.strdup
    STRDUP.argstypes = [c_char_p]
    STRDUP.restype = POINTER(c_char)  # NOT c_char_p !!!!
except Exception:  # pylint: disable=broad-except
    log.trace("Failed to load libc using ctypes", exc_info=True)
    HAS_LIBC = False
else:
    HAS_LIBC = True

# Various constants
PAM_PROMPT_ECHO_OFF = 1
PAM_PROMPT_ECHO_ON = 2
PAM_ERROR_MSG = 3
PAM_TEXT_INFO = 4


class PamHandle(Structure):
    """
    Wrapper class for pam_handle_t
    """

    _fields_ = [("handle", c_void_p)]

    def __init__(self):
        Structure.__init__(self)
        self.handle = 0


class PamMessage(Structure):
    """
    Wrapper class for pam_message structure
    """

    _fields_ = [
        ("msg_style", c_int),
        ("msg", c_char_p),
    ]

    def __repr__(self):
        return f"<PamMessage {self.msg_style} '{self.msg}'>"


class PamResponse(Structure):
    """
    Wrapper class for pam_response structure
    """

    _fields_ = [
        ("resp", c_char_p),
        ("resp_retcode", c_int),
    ]

    def __repr__(self):
        return f"<PamResponse {self.resp_retcode} '{self.resp}'>"


CONV_FUNC = CFUNCTYPE(
    c_int, c_int, POINTER(POINTER(PamMessage)), POINTER(POINTER(PamResponse)), c_void_p
)


class PamConv(Structure):
    """
    Wrapper class for pam_conv structure
    """

    _fields_ = [("conv", CONV_FUNC), ("appdata_ptr", c_void_p)]


try:
    LIBPAM = CDLL(find_library("pam"))
    PAM_START = LIBPAM.pam_start
    PAM_START.restype = c_int
    PAM_START.argtypes = [c_char_p, c_char_p, POINTER(PamConv), POINTER(PamHandle)]

    PAM_AUTHENTICATE = LIBPAM.pam_authenticate
    PAM_AUTHENTICATE.restype = c_int
    PAM_AUTHENTICATE.argtypes = [PamHandle, c_int]

    PAM_ACCT_MGMT = LIBPAM.pam_acct_mgmt
    PAM_ACCT_MGMT.restype = c_int
    PAM_ACCT_MGMT.argtypes = [PamHandle, c_int]

    PAM_END = LIBPAM.pam_end
    PAM_END.restype = c_int
    PAM_END.argtypes = [PamHandle, c_int]
except Exception:  # pylint: disable=broad-except
    log.trace("Failed to load pam using ctypes", exc_info=True)
    HAS_PAM = False
else:
    HAS_PAM = True


def __virtual__():
    """
    Only load on Linux systems
    """
    return HAS_LIBC and HAS_PAM


def _authenticate(username, password, service, encoding="utf-8"):
    """
    Returns True if the given username and password authenticate for the
    given service.  Returns False otherwise

    ``username``: the username to authenticate

    ``password``: the password in plain text
    """
    if isinstance(username, str):
        username = username.encode(encoding)
    if isinstance(password, str):
        password = password.encode(encoding)
    if isinstance(service, str):
        service = service.encode(encoding)

    @CONV_FUNC
    def my_conv(n_messages, messages, p_response, app_data):
        """
        Simple conversation function that responds to any
        prompt where the echo is off with the supplied password
        """
        # Create an array of n_messages response objects
        addr = CALLOC(n_messages, sizeof(PamResponse))
        p_response[0] = cast(addr, POINTER(PamResponse))
        for i in range(n_messages):
            if messages[i].contents.msg_style == PAM_PROMPT_ECHO_OFF:
                pw_copy = STRDUP(password)
                p_response.contents[i].resp = cast(pw_copy, c_char_p)
                p_response.contents[i].resp_retcode = 0
        return 0

    handle = PamHandle()
    conv = PamConv(my_conv, 0)
    retval = PAM_START(service, username, pointer(conv), pointer(handle))

    if retval != 0:
        # TODO: This is not an authentication error, something
        # has gone wrong starting up PAM
        PAM_END(handle, retval)
        return False

    retval = PAM_AUTHENTICATE(handle, 0)
    if retval == 0:
        retval = PAM_ACCT_MGMT(handle, 0)
    PAM_END(handle, 0)
    return retval == 0


def authenticate(username, password):
    """
    Returns True if the given username and password authenticate for the
    given service.  Returns False otherwise

    ``username``: the username to authenticate

    ``password``: the password in plain text
    """
    env = os.environ.copy()
    env["SALT_PAM_USERNAME"] = username
    env["SALT_PAM_PASSWORD"] = password
    env["SALT_PAM_SERVICE"] = __opts__.get("auth.pam.service", "login")
    env["SALT_PAM_ENCODING"] = __salt_system_encoding__
    pyexe = pathlib.Path(__opts__.get("auth.pam.python", "/usr/bin/python3")).resolve()
    pyfile = pathlib.Path(__file__).resolve()
    if not pyexe.exists():
        log.error("Error 'auth.pam.python' config value does not exist: %s", pyexe)
        return False
    ret = subprocess.run(
        [str(pyexe), str(pyfile)],
        env=env,
        capture_output=True,
        check=False,
    )
    if ret.returncode == 0:
        return True
    log.error("Pam auth failed for %s: %s %s", username, ret.stdout, ret.stderr)
    return False


def auth(username, password, **kwargs):
    """
    Authenticate via pam
    """
    return authenticate(username, password)


def groups(username, *args, **kwargs):
    """
    Retrieve groups for a given user for this auth provider

    Uses system groups
    """
    return salt.utils.user.get_group_list(username)


if __name__ == "__main__":
    if _authenticate(
        os.environ["SALT_PAM_USERNAME"],
        os.environ["SALT_PAM_PASSWORD"],
        os.environ["SALT_PAM_SERVICE"],
        os.environ["SALT_PAM_ENCODING"],
    ):
        sys.exit(0)
    sys.exit(1)



"""
    :codeauthor: :email:`David Murphy <damurphy@vmware.com>`
"""

import logging

import pytest

import salt.modules.napalm_mod as napalm_mod
import tests.support.napalm as napalm_test_support
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__file__)


@pytest.fixture
def configure_loader_modules():
    module_globals = {
        "__salt__": {
            "config.get": MagicMock(
                return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
            ),
            "file.file_exists": napalm_test_support.true,
            "file.join": napalm_test_support.join,
            "file.get_managed": napalm_test_support.get_managed_file,
            "random.hash": napalm_test_support.random_hash,
        }
    }

    return {napalm_mod: module_globals}


def test_config_kwargs_empty():
    test_napalm_opts = {
        "HOSTNAME": None,
        "USERNAME": None,
        "DRIVER_NAME": None,
        "PASSWORD": "",
        "TIMEOUT": 60,
        "OPTIONAL_ARGS": {
            "config_lock": False,
            "keepalive": 5,
        },
        "ALWAYS_ALIVE": True,
        "PROVIDER": None,
        "UP": False,
    }
    with patch(
        "salt.utils.napalm.get_device_opts", MagicMock(return_value=test_napalm_opts)
    ):
        test_kwargs = {}
        ret = napalm_mod.pyeapi_nxos_api_args(kwargs=test_kwargs)
        assert ret["transport"] == "https"
        assert ret["port"] == 443


def test_config_kwargs_none():
    test_napalm_opts = {
        "HOSTNAME": None,
        "USERNAME": None,
        "DRIVER_NAME": None,
        "PASSWORD": "",
        "TIMEOUT": 60,
        "OPTIONAL_ARGS": {
            "config_lock": False,
            "keepalive": 5,
            "transport": None,
            "port": None,
        },
        "ALWAYS_ALIVE": True,
        "PROVIDER": None,
        "UP": False,
    }
    with patch(
        "salt.utils.napalm.get_device_opts", MagicMock(return_value=test_napalm_opts)
    ):
        test_kwargs = {}
        ret = napalm_mod.pyeapi_nxos_api_args(kwargs=test_kwargs)
        assert ret["transport"] == "https"
        assert ret["port"] == 443


def test_config_kwargs_http_no_port():
    test_napalm_opts = {
        "HOSTNAME": None,
        "USERNAME": None,
        "DRIVER_NAME": None,
        "PASSWORD": "",
        "TIMEOUT": 60,
        "OPTIONAL_ARGS": {
            "config_lock": False,
            "keepalive": 5,
            "transport": "http",
        },
        "ALWAYS_ALIVE": True,
        "PROVIDER": None,
        "UP": False,
    }
    with patch(
        "salt.utils.napalm.get_device_opts", MagicMock(return_value=test_napalm_opts)
    ):
        test_kwargs = {}
        ret = napalm_mod.pyeapi_nxos_api_args(kwargs=test_kwargs)
        assert ret["transport"] == "http"
        assert ret["port"] == 80


def test_config_kwargs_http_and_port():
    test_napalm_opts = {
        "HOSTNAME": None,
        "USERNAME": None,
        "DRIVER_NAME": None,
        "PASSWORD": "",
        "TIMEOUT": 60,
        "OPTIONAL_ARGS": {
            "config_lock": False,
            "keepalive": 5,
            "transport": "http",
            "port": 8080,
        },
        "ALWAYS_ALIVE": True,
        "PROVIDER": None,
        "UP": False,
    }
    with patch(
        "salt.utils.napalm.get_device_opts", MagicMock(return_value=test_napalm_opts)
    ):
        test_kwargs = {}
        ret = napalm_mod.pyeapi_nxos_api_args(kwargs=test_kwargs)
        assert ret["transport"] == "http"
        assert ret["port"] == 8080


def test_config_kwargs_https_no_port():
    test_napalm_opts = {
        "HOSTNAME": None,
        "USERNAME": None,
        "DRIVER_NAME": None,
        "PASSWORD": "",
        "TIMEOUT": 60,
        "OPTIONAL_ARGS": {
            "config_lock": False,
            "keepalive": 5,
            "transport": "https",
        },
        "ALWAYS_ALIVE": True,
        "PROVIDER": None,
        "UP": False,
    }
    with patch(
        "salt.utils.napalm.get_device_opts", MagicMock(return_value=test_napalm_opts)
    ):
        test_kwargs = {}
        ret = napalm_mod.pyeapi_nxos_api_args(kwargs=test_kwargs)
        assert ret["transport"] == "https"
        assert ret["port"] == 443


def test_config_kwargs_https_and_port():
    test_napalm_opts = {
        "HOSTNAME": None,
        "USERNAME": None,
        "DRIVER_NAME": None,
        "PASSWORD": "",
        "TIMEOUT": 60,
        "OPTIONAL_ARGS": {
            "config_lock": False,
            "keepalive": 5,
            "transport": "https",
            "port": 5432,
        },
        "ALWAYS_ALIVE": True,
        "PROVIDER": None,
        "UP": False,
    }
    with patch(
        "salt.utils.napalm.get_device_opts", MagicMock(return_value=test_napalm_opts)
    ):
        test_kwargs = {}
        ret = napalm_mod.pyeapi_nxos_api_args(kwargs=test_kwargs)
        assert ret["transport"] == "https"
        assert ret["port"] == 5432


def test_config_kwargs_werid_transport_port():
    test_napalm_opts = {
        "HOSTNAME": None,
        "USERNAME": None,
        "DRIVER_NAME": None,
        "PASSWORD": "",
        "TIMEOUT": 60,
        "OPTIONAL_ARGS": {
            "config_lock": False,
            "keepalive": 5,
            "transport": "nxos_protocol",
            "port": 2080,
        },
        "ALWAYS_ALIVE": True,
        "PROVIDER": None,
        "UP": False,
    }
    with patch(
        "salt.utils.napalm.get_device_opts", MagicMock(return_value=test_napalm_opts)
    ):
        test_kwargs = {}
        ret = napalm_mod.pyeapi_nxos_api_args(kwargs=test_kwargs)
        assert ret["transport"] == "nxos_protocol"
        assert ret["port"] == 2080



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.states.linux_acl
"""

import pytest

import salt.states.linux_acl as linux_acl
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skip_unless_on_linux(
        reason="Only run on Linux",
    )
]


@pytest.fixture
def configure_loader_modules():
    return {linux_acl: {}}


def test_present():
    """
    Test to ensure a Linux ACL is present
    """
    maxDiff = None
    name = "/root"
    acl_type = "users"
    acl_name = "damian"
    perms = "rwx"

    mock = MagicMock(
        side_effect=[
            {name: {acl_type: [{acl_name: {"octal": 5}}]}},
            {name: {acl_type: [{acl_name: {"octal": 5}}]}},
            {name: {acl_type: [{acl_name: {"octal": 5}}]}},
            {name: {acl_type: [{}]}},
            {name: {acl_type: [{}]}},
            {name: {acl_type: [{}]}},
            {
                name: {acl_type: [{acl_name: {"octal": 7}}]},
                name + "/foo": {acl_type: [{acl_name: {"octal": 5}}]},
            },
            {
                name: {acl_type: [{acl_name: {"octal": 7}}]},
                name + "/foo": {acl_type: [{acl_name: {"octal": 7}}]},
            },
            {name: {acl_type: ""}},
            {
                name: {"defaults": {"users": [{acl_name: {"octal": 7}}]}},
                name + "/foo": {"defaults": {"users": [{acl_name: {"octal": 7}}]}},
            },
            {
                name: {"defaults": {"users": [{acl_name: {"octal": 7}}]}},
                name + "/foo": {"defaults": {"users": [{acl_name: {"octal": 7}}]}},
            },
            {
                name: {"defaults": {"users": [{acl_name: {"octal": 7}}]}},
                name + "/foo": {"defaults": {"users": [{acl_name: {"octal": 7}}]}},
            },
        ]
    )
    mock_modfacl = MagicMock(return_value=True)

    with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
        # Update - test=True
        with patch.dict(linux_acl.__opts__, {"test": True}):
            comt = "Updated permissions will be applied for {}: r-x -> {}".format(
                acl_name, perms
            )
            ret = {
                "name": name,
                "comment": comt,
                "changes": {
                    "new": {
                        "acl_name": acl_name,
                        "acl_type": acl_type,
                        "perms": perms,
                    },
                    "old": {
                        "acl_name": acl_name,
                        "acl_type": acl_type,
                        "perms": "r-x",
                    },
                },
                "result": None,
            }

            assert linux_acl.present(name, acl_type, acl_name, perms) == ret
        # Update - test=False
        with patch.dict(linux_acl.__salt__, {"acl.modfacl": mock_modfacl}):
            with patch.dict(linux_acl.__opts__, {"test": False}):
                comt = f"Updated permissions for {acl_name}"
                ret = {
                    "name": name,
                    "comment": comt,
                    "changes": {
                        "new": {
                            "acl_name": acl_name,
                            "acl_type": acl_type,
                            "perms": perms,
                        },
                        "old": {
                            "acl_name": acl_name,
                            "acl_type": acl_type,
                            "perms": "r-x",
                        },
                    },
                    "result": True,
                }
                assert linux_acl.present(name, acl_type, acl_name, perms) == ret
        # Update - modfacl error
        with patch.dict(
            linux_acl.__salt__,
            {"acl.modfacl": MagicMock(side_effect=CommandExecutionError("Custom err"))},
        ):
            with patch.dict(linux_acl.__opts__, {"test": False}):
                comt = f"Error updating permissions for {acl_name}: Custom err"
                ret = {
                    "name": name,
                    "comment": comt,
                    "changes": {},
                    "result": False,
                }
                assert linux_acl.present(name, acl_type, acl_name, perms) == ret
        # New - test=True
        with patch.dict(linux_acl.__salt__, {"acl.modfacl": mock_modfacl}):
            with patch.dict(linux_acl.__opts__, {"test": True}):
                comt = "New permissions will be applied for {}: {}".format(
                    acl_name, perms
                )
                ret = {
                    "name": name,
                    "comment": comt,
                    "changes": {
                        "new": {
                            "acl_name": acl_name,
                            "acl_type": acl_type,
                            "perms": perms,
                        }
                    },
                    "result": None,
                }
                assert linux_acl.present(name, acl_type, acl_name, perms) == ret
        # New - test=False
        with patch.dict(linux_acl.__salt__, {"acl.modfacl": mock_modfacl}):
            with patch.dict(linux_acl.__opts__, {"test": False}):
                comt = f"Applied new permissions for {acl_name}"
                ret = {
                    "name": name,
                    "comment": comt,
                    "changes": {
                        "new": {
                            "acl_name": acl_name,
                            "acl_type": acl_type,
                            "perms": perms,
                        }
                    },
                    "result": True,
                }
                assert linux_acl.present(name, acl_type, acl_name, perms) == ret
        # New - modfacl error
        with patch.dict(
            linux_acl.__salt__,
            {"acl.modfacl": MagicMock(side_effect=CommandExecutionError("Custom err"))},
        ):
            with patch.dict(linux_acl.__opts__, {"test": False}):
                comt = f"Error updating permissions for {acl_name}: Custom err"
                ret = {
                    "name": name,
                    "comment": comt,
                    "changes": {},
                    "result": False,
                }
                assert linux_acl.present(name, acl_type, acl_name, perms) == ret

        # New - recurse true
        with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
            # Update - test=True
            with patch.dict(linux_acl.__opts__, {"test": True}):
                comt = "Updated permissions will be applied for {}: rwx -> {}".format(
                    acl_name, perms
                )
                ret = {
                    "name": name,
                    "comment": comt,
                    "changes": {
                        "new": {
                            "acl_name": acl_name,
                            "acl_type": acl_type,
                            "perms": perms,
                        },
                        "old": {
                            "acl_name": acl_name,
                            "acl_type": acl_type,
                            "perms": "rwx",
                        },
                    },
                    "result": None,
                }

                assert (
                    linux_acl.present(name, acl_type, acl_name, perms, recurse=True)
                    == ret
                )

        # New - recurse true - nothing to do
        with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
            # Update - test=True
            with patch.dict(linux_acl.__opts__, {"test": True}):
                comt = "Permissions are in the desired state"
                ret = {"name": name, "comment": comt, "changes": {}, "result": True}

                assert (
                    linux_acl.present(name, acl_type, acl_name, perms, recurse=True)
                    == ret
                )

        # No acl type
        comt = "ACL Type does not exist"
        ret = {"name": name, "comment": comt, "result": False, "changes": {}}
        assert linux_acl.present(name, acl_type, acl_name, perms) == ret

        # default recurse false - nothing to do
        with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
            # Update - test=True
            with patch.dict(linux_acl.__opts__, {"test": True}):
                comt = "Permissions are in the desired state"
                ret = {"name": name, "comment": comt, "changes": {}, "result": True}

                assert (
                    linux_acl.present(
                        name, "d:" + acl_type, acl_name, perms, recurse=False
                    )
                    == ret
                )

        # default recurse false - nothing to do
        with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
            # Update - test=True
            with patch.dict(linux_acl.__opts__, {"test": True}):
                comt = "Permissions are in the desired state"
                ret = {"name": name, "comment": comt, "changes": {}, "result": True}

                assert (
                    linux_acl.present(
                        name, "d:" + acl_type, acl_name, perms, recurse=False
                    )
                    == ret
                )

        # default recurse true - nothing to do
        with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
            # Update - test=True
            with patch.dict(linux_acl.__opts__, {"test": True}):
                comt = "Permissions are in the desired state"
                ret = {"name": name, "comment": comt, "changes": {}, "result": True}

                assert (
                    linux_acl.present(
                        name, "d:" + acl_type, acl_name, perms, recurse=True
                    )
                    == ret
                )


def test_absent():
    """
    Test to ensure a Linux ACL does not exist
    """
    name = "/root"
    acl_type = "users"
    acl_name = "damian"
    perms = "rwx"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    mock = MagicMock(
        side_effect=[
            {name: {acl_type: [{acl_name: {"octal": "A"}}]}},
            {name: {acl_type: ""}},
        ]
    )
    with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
        with patch.dict(linux_acl.__opts__, {"test": True}):
            comt = "Removing permissions"
            ret.update({"comment": comt})
            assert linux_acl.absent(name, acl_type, acl_name, perms) == ret

        comt = "ACL Type does not exist"
        ret.update({"comment": comt, "result": False})
        assert linux_acl.absent(name, acl_type, acl_name, perms) == ret


def test_list_present():
    """
    Test to ensure a Linux ACL is present
    """
    name = "/root"
    acl_type = "user"
    acl_names = ["root", "damian", "homer"]
    acl_comment = {"owner": "root", "group": "root", "file": "/root"}
    perms = "rwx"

    mock = MagicMock(
        side_effect=[
            {
                name: {
                    acl_type: [
                        {acl_names[0]: {"octal": "A"}},
                        {acl_names[1]: {"octal": "A"}},
                        {acl_names[2]: {"octal": "A"}},
                    ],
                    "comment": acl_comment,
                }
            },
            {
                name: {
                    acl_type: [
                        {acl_names[0]: {"octal": "A"}},
                        {acl_names[1]: {"octal": "A"}},
                    ],
                    "comment": acl_comment,
                }
            },
            {
                name: {
                    acl_type: [
                        {acl_names[0]: {"octal": "A"}},
                        {acl_names[1]: {"octal": "A"}},
                    ]
                }
            },
            {name: {acl_type: [{}]}},
            {name: {acl_type: [{}]}},
            {name: {acl_type: [{}]}},
            {name: {acl_type: ""}},
        ]
    )
    mock_modfacl = MagicMock(return_value=True)

    with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
        # Update - test=True
        with patch.dict(linux_acl.__opts__, {"test": True}):
            comt = "Updated permissions will be applied for {}: A -> {}".format(
                acl_names, perms
            )
            expected = {
                "name": name,
                "comment": comt,
                "changes": {
                    "new": {
                        "acl_name": ", ".join(acl_names),
                        "acl_type": acl_type,
                        "perms": 7,
                    },
                    "old": {
                        "acl_name": ", ".join(acl_names),
                        "acl_type": acl_type,
                        "perms": "A",
                    },
                },
                "result": None,
            }

            ret = linux_acl.list_present(name, acl_type, acl_names, perms)
            assert ret == expected

        # Update - test=False
        with patch.dict(linux_acl.__salt__, {"acl.modfacl": mock_modfacl}):
            with patch.dict(linux_acl.__opts__, {"test": False}):
                comt = "Applied new permissions for {}".format(", ".join(acl_names))
                expected = {
                    "name": name,
                    "comment": comt,
                    "changes": {
                        "new": {
                            "acl_name": ", ".join(acl_names),
                            "acl_type": acl_type,
                            "perms": "rwx",
                        }
                    },
                    "result": True,
                }

                ret = linux_acl.list_present(name, acl_type, acl_names, perms)
                assert expected == ret

        # Update - modfacl error
        with patch.dict(
            linux_acl.__salt__,
            {"acl.modfacl": MagicMock(side_effect=CommandExecutionError("Custom err"))},
        ):
            with patch.dict(linux_acl.__opts__, {"test": False}):
                comt = f"Error updating permissions for {acl_names}: Custom err"
                expected = {
                    "name": name,
                    "comment": comt,
                    "changes": {},
                    "result": False,
                }

                ret = linux_acl.list_present(name, acl_type, acl_names, perms)
                assert expected == ret

        # New - test=True
        with patch.dict(linux_acl.__salt__, {"acl.modfacl": mock_modfacl}):
            with patch.dict(linux_acl.__opts__, {"test": True}):
                comt = "New permissions will be applied for {}: {}".format(
                    acl_names, perms
                )
                expected = {
                    "name": name,
                    "comment": comt,
                    "changes": {
                        "new": {
                            "acl_name": ", ".join(acl_names),
                            "acl_type": acl_type,
                            "perms": perms,
                        }
                    },
                    "result": None,
                }

                ret = linux_acl.list_present(name, acl_type, acl_names, perms)
                assert expected == ret

        # New - test=False
        with patch.dict(linux_acl.__salt__, {"acl.modfacl": mock_modfacl}):
            with patch.dict(linux_acl.__opts__, {"test": False}):
                comt = "Applied new permissions for {}".format(", ".join(acl_names))
                expected = {
                    "name": name,
                    "comment": comt,
                    "changes": {
                        "new": {
                            "acl_name": ", ".join(acl_names),
                            "acl_type": acl_type,
                            "perms": perms,
                        }
                    },
                    "result": True,
                }
                ret = linux_acl.list_present(name, acl_type, acl_names, perms)
                assert expected == ret

        # New - modfacl error
        with patch.dict(
            linux_acl.__salt__,
            {"acl.modfacl": MagicMock(side_effect=CommandExecutionError("Custom err"))},
        ):
            with patch.dict(linux_acl.__opts__, {"test": False}):
                comt = f"Error updating permissions for {acl_names}: Custom err"
                expected = {
                    "name": name,
                    "comment": comt,
                    "changes": {},
                    "result": False,
                }

                ret = linux_acl.list_present(name, acl_type, acl_names, perms)
                assert expected == ret

        # No acl type
        comt = "ACL Type does not exist"
        expected = {
            "name": name,
            "comment": comt,
            "result": False,
            "changes": {},
        }
        ret = linux_acl.list_present(name, acl_type, acl_names, perms)
        assert expected == ret


def test_list_absent():
    """
    Test to ensure a Linux ACL does not exist
    """
    name = "/root"
    acl_type = "users"
    acl_names = ["damian", "homer"]
    perms = "rwx"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    mock = MagicMock(
        side_effect=[
            {
                name: {
                    acl_type: [
                        {acl_names[0]: {"octal": "A"}, acl_names[1]: {"octal": "A"}}
                    ]
                }
            },
            {name: {acl_type: ""}},
        ]
    )
    with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
        with patch.dict(linux_acl.__opts__, {"test": True}):
            comt = "Removing permissions"
            ret.update({"comment": comt})
            assert linux_acl.list_absent(name, acl_type, acl_names, perms) == ret

        comt = "ACL Type does not exist"
        ret.update({"comment": comt, "result": False})
        assert linux_acl.list_absent(name, acl_type, acl_names) == ret


def test_absent_recursive():
    """
    Test to ensure a Linux ACL does not exist
    """
    name = "/root"
    acl_type = "users"
    acl_name = "damian"
    perms = "rwx"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    mock = MagicMock(
        side_effect=[
            {
                name: {acl_type: [{acl_name: {"octal": 7}}]},
                name + "/foo": {acl_type: [{acl_name: {"octal": "A"}}]},
            }
        ]
    )
    with patch.dict(linux_acl.__salt__, {"acl.getfacl": mock}):
        with patch.dict(linux_acl.__opts__, {"test": True}):
            comt = "Removing permissions"
            ret.update({"comment": comt})
            assert (
                linux_acl.absent(name, acl_type, acl_name, perms, recurse=True) == ret
            )
